/*
QKD.cpp
Written by Walter O. Krawec

Copyright (c) 2016 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "QKD.h"

namespace QKDEvolve
{

void EvolveSystem::rollOver(int& t, int& b, int& a, int& e, int& p)
{
	++p;
	if(p >= 2)
	{
		p = 0;
		++e;
	}
	if(e >= 4)
	{
		e = 0;
		++a;
	}
	if(a >= 2)
	{
		a = 0;
		++b;
	}
	if(b >= 2)
	{
		b = 0;
		++t;
	}
}


// traces out first 2 dim subsystem
void EvolveSystem::fastTrace(algebra::mat& input, algebra::mat& output)
{
	output.create(input.getRows()/2);
	for(int i=0; i<output.getRows(); ++i)
	{
		for(int j=0; j<output.getCols(); ++j)
		{
			output(i,j) = input(i,j) + input(i+input.getRows()/2,j+input.getRows()/2);
		}
	}
}

void EvolveSystem::evolve(QKD* protocol, double targetQBER, std::vector <quantum::Basis>& output)
{
  int numGenerations = 250;
	std::ofstream statFile;
	statFile.open("stats.txt");
	statFile << "best\tavg\tQ\tQ-best\tR\tother\n";
	statFile.close();
	const int CROSSOVER_AMOUNT=4;
	int startPopulation = 50;
	int maxPopulation = 50;
	quantum::DensityList P2;
	std::list <CS> population;
	std::cout << "Setting up initial population...\n";
	double smallScale=0.25, largeScale=0.25;
	smallScale = 1;
	largeScale = 1;

	//smallScale = .1;
	//largeScale = .1;

	for(int i=0; i<startPopulation; ++i)
	{
		if(i%5 == 0)
			std::cout << i << "/" << startPopulation << "\n";
		CS cs;
		cs.setup(protocol, largeScale, smallScale);
		cs.calculateFitness(protocol, targetQBER);

		//protocol->testSolution(cs.E);

		population.push_back(cs);
		//return;
	}
	//return;

	population.sort();
	std::cout << "DEBUG: max fitness = " << population.begin()->fitness << "\n";
	std::cout << "DEBUG: min fitness = " << population.rbegin()->fitness << "\n";

	//std::cin.get();

	std::list <CS>::iterator Iter, Iter2;
	int index = 0;

	double bestQBER = 1;

	for(int t=0; t<numGenerations; ++t)
	{
#ifdef WIN32
		if(_kbhit()) break;
#endif
		int startingSize = (int)population.size();
		//for(int i=0; i<startingSize / CROSSOVER_AMOUNT + 1; ++i)
		for(int i=0; i<2*startingSize/4+1; ++i)
		{
#ifdef WIN32
			if(_kbhit()) break;
#endif
			//std::cout << i << "/" << startingSize / CROSSOVER_AMOUNT+1 << "\n";
			int r = rand()%(startingSize / CROSSOVER_AMOUNT + 1);
			Iter = population.begin();
			for(int j=0; j<r; ++j)
			{
				++Iter;
				if(Iter == population.end())
				{
					Iter = population.begin();
					break;
				}
			}
			CS child;
			r = rand()%startingSize;//(startingSize / CROSSOVER_AMOUNT + 1);
			Iter2 = population.begin();
			for(int j=0; j<r; ++j)
			{
				++Iter2;
				if(Iter2 == population.end())
				{
					Iter2 = population.begin();
					break;
				}
			}

			Iter->crossOver(*Iter2, child);
			//calculateFitness(child, rules);
			child.calculateFitness(protocol, targetQBER);
			population.push_back(child);
		}

		Iter = population.begin();
		++Iter;
		for(; Iter != population.end(); ++Iter)
		{
#ifdef WIN32
			if(_kbhit())break;
#endif
			if(rand()%100 < 40)
			{
				Iter->mutate();
				Iter->calculateFitness(protocol, targetQBER);
			}
		}

		population.sort();

		// 2?
		for(int i=0; i<1*startingSize/CROSSOVER_AMOUNT +1; ++i)
		{//break;
#ifdef WIN32
			if(_kbhit()) break;
#endif
			//std::cout << i << "/" << startingSize / CROSSOVER_AMOUNT +1 << ": " << population.begin()->fitness << " -> ";
			CS child;
			//population.begin()->randomEvolve(child);
			child = *population.begin();
			child.mutate();
			child.calculateFitness(protocol, targetQBER);
			//std::cout << child.fitness << "\n";
			population.push_back(child);
		}

		population.sort();
		int endingSize = (int)population.size();
		while(endingSize > maxPopulation)
		{
			population.pop_back();
			--endingSize;
		}
		//if(index%20 == 0)
			population.begin()->rateFit = protocol->computeKeyRate(population.begin()->E, false);
		double rateFit = population.begin()->rateFit;

		double avgFit = 0;
		for(Iter=population.begin(); Iter != population.end(); ++Iter)
			avgFit += Iter->fitness;
		avgFit = avgFit / (double)population.size();

		statFile.open("stats.txt", std::ios_base::app);
		statFile  << population.begin()->fitness << "\t" << avgFit << "\t" << bestQBER;

		statFile << "\t" << population.begin()->Q << "\t" << rateFit;

		for(int i=0; i<(int)population.begin()->qberList.size(); ++i)
			statFile << "\t" << population.begin()->qberList[i];

		population.begin()->calculateFitness(protocol, targetQBER);

		protocol->printDebugInfo(statFile, true);//(index%20==0));

		statFile << "\n";
		statFile.close();

		++index;
		//population.begin()->calculateFitness(protocol, targetQBER);

		std::cout << "DEBUG: fitness = " << population.begin()->fitness << ", qberFit = " << population.begin()->qberFit << ", Q = " << population.begin()->Q << ",\n\trate = " << rateFit << "\n";
		for(int i=0; i<(int)population.begin()->qberList.size(); ++i)
			std::cout << "\t" << i << ": " << population.begin()->qberList[i] << "\n";

		if(rateFit < 0 && population.begin()->Q < bestQBER)
			bestQBER = population.begin()->Q;

		std::cout << "\nOUTPUT: Q = " << bestQBER << "\n\n";


		//std::cout << "DEBUG: fitness = " << population.begin()->fitness << ", qberFit = " << population.begin()->qberFit << ", Q = " << population.begin()->Q << ",\n\trateFit = " << rateFit << ", E = " << population.begin()->EFit << "\n";
		//std::cout << "\tH1 = " << population.begin()->H1 << ", H2 = " << population.begin()->H2 << "\n";
	}

	population.sort();

	population.begin()->rateFit = protocol->computeKeyRate(population.begin()->E, true);
	++index;
	std::cout << "DEBUG: fitness = " << population.begin()->fitness << ", qberFit = " << population.begin()->qberFit << ", Q = " << population.begin()->Q << ", rateFit = " << population.begin()->rateFit << ", E = " << population.begin()->EFit << "\n";
	for(int i=0; i<(int)population.begin()->qberList.size(); ++i)
		std::cout << "\t" << i << ": " << population.begin()->qberList[i] << "\n";
	output = population.begin()->E;
	population.begin()->saveMathematica("output.txt");

	population.begin()->saveLocalDim1("output2.txt");
	return;
}
}
