/*
QKDEvolve
Written by Walter O. Krawec

Copyright (c) 2016 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <iostream>
#include <time.h>

#include "Quantum/Quantum.h"
#include "QKD.h"

#include "Examples/PLProtocol.h"
#include "PL/PL.h"

int main(int argc, char** argv)
{
  if(argc != 3)
    {
      std::cout << "QKDEvolve: A GA to analyze the security of QKD protocols\n\n";
      std::cout << "Usage: QKDEvolve [protocol-file] [target-QBER]\n";
      std::cout << "\t[protocol-file] is the protocol to analyze.\n";
      std::cout << "\t[target-QBER] is the target QBER.\n\n";
      return 0;
    }
	srand((unsigned int)time(NULL));

	pl::PL pl1;
	quantum::DensityList dl;
	std::list <quantum::DensityList> qberList, statList, printList;
	std::vector <int> dimE;

	pl1.constructDL(argv[1], dl, qberList, statList, printList, dimE);


	PLProtocol plProt;
	plProt.setup(dl, qberList, statList, printList, dimE);
	QKDEvolve::EvolveSystem es;
	std::vector <quantum::Basis> output;

	es.evolve(&plProt, atof(argv[2]), output);


	return 0;
}
